/* 
 * Exemple d'clairage avec dessin de sphere.
 * Auteur : Fares BELHADJ (Lire COPYING)
 * Email  : amsi@ai.univ-paris8.fr
 * Date   : 22/10/2007
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <SDL.h>
#include <GL/gl.h>
#include <GL/glu.h>

static SDL_Surface * sdlInit(int w, int h);
static void          bgr2rgb(unsigned char * data, int size);
static void          sdlInitGL(SDL_Surface * s);
static void          sdlResizeGL(SDL_Surface * s);
static void          sdlLoop(SDL_Surface * s);
static int           sdlManageEvents(SDL_Surface * s);
static void          sdlDraw(SDL_Surface * s);
static void          cube(void);
static inline void   sdlFPS(void);

static int pause = 0, degres = 360;

static GLuint texId[3] = {0, 0, 0}, curId = 0;

static SDL_Surface * texSurface = NULL;

/* un tout petit main */
int main(int argc, char ** argv) {
  SDL_Surface * s;
  s = sdlInit(640, 480);
  sdlLoop(s);
  return 0;
}

/* sdlInit : initialiser SDL pour OpenGL */
static SDL_Surface * sdlInit(int w, int h) {
  SDL_Surface * s;
  const SDL_VideoInfo * vi = NULL;
  if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
    fprintf(stderr, "Impossible d'initialiser SDL: %s\n", SDL_GetError());
    exit(1);
  }

  if( (vi = SDL_GetVideoInfo()) == 0 ) {
    fprintf(stderr, "Echec de SDL_GetVideoInfo(): %s\n", SDL_GetError());
    exit(1);
  }
  SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
  SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
  SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
  SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
  SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
  if ( (s = SDL_SetVideoMode(w, h, vi->vfmt->BitsPerPixel,
			     SDL_OPENGL | SDL_RESIZABLE)) == NULL ) {
    fprintf(stderr, "Impossible d'ouvrir le mode video : %s\n", SDL_GetError());
    exit(1);
  }
  atexit(SDL_Quit);
  sdlInitGL(s);
  return s;
}

static void bgr2rgb(unsigned char * data, int size) {
  int i;
  unsigned char c;
  for (i = 0; i < size; i += 3) {
      c = data[i];
      data[i] = data[i + 2];
      data[i + 2] = c;
  }
}

/* sdlInitGL : configurer les options GL */
static void sdlInitGL(SDL_Surface * s) {
  glShadeModel(GL_SMOOTH);
  glDepthFunc(GL_LEQUAL);
  glFrontFace(GL_CCW);
  glCullFace(GL_BACK);
  glClearColor(0, 0, 0, 0);
  glClearDepth(1.0);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  glDisable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);

  glPolygonMode(GL_FRONT, GL_FILL);
  glPolygonMode(GL_BACK, GL_LINE);

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  
  if( (texSurface = SDL_LoadBMP("tex01.bmp")) == NULL ) {
    fprintf(stderr, "Impossible d'ouvrir le fichier : %s\n", SDL_GetError());
    exit(1);
  }
  bgr2rgb(texSurface->pixels, 3 * texSurface->w * texSurface->h);

  glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T, GL_REPEAT);

  glGenTextures(3, texId);

  glBindTexture(GL_TEXTURE_2D, texId[0]); 
  glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexImage2D(GL_TEXTURE_2D, 0, 3, texSurface->w, texSurface->h, 0, GL_RGB, GL_UNSIGNED_BYTE, texSurface->pixels);

  glBindTexture(GL_TEXTURE_2D, texId[1]); 
  glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexImage2D(GL_TEXTURE_2D, 0, 3, texSurface->w, texSurface->h, 0, GL_RGB, GL_UNSIGNED_BYTE, texSurface->pixels);
  
  glBindTexture(GL_TEXTURE_2D, texId[2]); 
  glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
  glTexImage2D(GL_TEXTURE_2D, 0, 3, texSurface->w, texSurface->h, 0, GL_RGB, GL_UNSIGNED_BYTE, texSurface->pixels);
  gluBuild2DMipmaps(GL_TEXTURE_2D, 3, texSurface->w, texSurface->h, GL_RGB, GL_UNSIGNED_BYTE, texSurface->pixels);
  
  glEnable(GL_TEXTURE_2D);

  SDL_FreeSurface(texSurface);

  sdlResizeGL(s);

}

/* 
 * CallBack : en cas de resize modifier la matrice de projection en
 * fonction du rapport w/h
 */
static void sdlResizeGL(SDL_Surface * s) {
  glViewport(0, 0, s->w, s->h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(45.0, s->w / (float)s->h, 0.1, 100.0);
  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity( );
}

/* 
 * loop principale : sdlManageEvents renvoie vrai si la fenetre est
 * inactive => tantque inactive dors une milliseconde.
 * sinon dessiner puis afficher le framerate
 */
static void sdlLoop(SDL_Surface * s) {
  for(;;) {
    while(sdlManageEvents(s)) SDL_Delay(1);
    sdlDraw(s);
    sdlFPS();
  }
}

/* 
 * sdlManageEvents gere les evenements ; aussi renvoie vrai si la
 * fenetre est inactive, sinon faux.
 * Les evenements sont : 
 *
 * escape, q, Q : quitter
 *
 * p : pause
 *
 * m : changer de filtrage pour la texture
 *
 * w : activer / desactiver le mode wireframe
 *
 * l : activer / desactiver l'eclairage
 *
 * f : activer / desactiver le dgrad
 *
 * F1 : fullscreen On / Off
 * 
 * redimensionnement de fenetre : appeler sdlResizeGL
 *
 */
static int sdlManageEvents(SDL_Surface * s) {
  GLint v[2];
  static int active = 1;
  SDL_Event event;
  while(SDL_PollEvent(&event)) 
    switch (event.type) {
    case SDL_KEYDOWN:
      switch(event.key.keysym.sym) {
      case SDLK_LEFT:
	degres -= 10;
	break;
      case SDLK_RIGHT:
	degres += 10;
	break;
      case ' ':
      case 'p':
	pause = !pause;
	break;
      case 'm':
	curId = (curId + 1) % 3;
	break;
      case 'w':
	glGetIntegerv(GL_POLYGON_MODE, v);
	if(v[0] == GL_FILL)
	  glPolygonMode(GL_FRONT, GL_LINE);
	else
	  glPolygonMode(GL_FRONT, GL_FILL);
	break;
      case 'f':
	glGetIntegerv(GL_SHADE_MODEL, v);
	if(v[0] == GL_FLAT)
	  glShadeModel(GL_SMOOTH);
	else
	  glShadeModel(GL_FLAT);
	break;
      case 'l':
	if(glIsEnabled(GL_LIGHTING)) {
	  glDisable(GL_LIGHTING);
	  glDisable(GL_LIGHT0);
	} else {
	  glEnable(GL_LIGHTING);
	  glEnable(GL_LIGHT0);
	}
	break;
      case SDLK_ESCAPE:
      case 'q':
	exit(0);
      case SDLK_F1:
	SDL_WM_ToggleFullScreen(s);
	break;
      default:
	fprintf(stderr, "La touche %s a ete pressee\n",
		SDL_GetKeyName(event.key.keysym.sym));
	break;
      }
      break;
    case SDL_KEYUP:
      break;
    case SDL_ACTIVEEVENT:
      if(event.active.state & SDL_APPACTIVE)
	active = event.active.gain;
      break;
    case SDL_VIDEORESIZE:
      if( (s = SDL_SetVideoMode(event.resize.w, event.resize.h, 
				s->format->BitsPerPixel, s->flags)) == NULL ) {
	fprintf(stderr, "Echec de SDL_SetVideoMode apres le resize : %s\n", 
		SDL_GetError());
	exit(1);
      }
      sdlResizeGL(s);
      break;
    case SDL_QUIT:
      exit(0);
    }
  return !active;
}

static void sdlDraw(SDL_Surface * s) {
  Uint32 dt;
  int xm, ym;
  static float a = 0.0;
  static Uint32 t0 = 0;
  dt = SDL_GetTicks() - t0;
  t0 += dt;

  SDL_PumpEvents();
  SDL_GetMouseState(&xm, &ym);

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glLoadIdentity();

  glBindTexture(GL_TEXTURE_2D, texId[curId]);

  gluLookAt(0, 4.0 * sin(a * M_PI / 180.0), -5, 5.0 * (xm - (s->w >> 1)) / (float)(s->w),  5.0 * (ym - (s->h >> 1)) / (float)(s->h), 0, 0, 1, 0);

  glPushMatrix();
  glTranslatef(1.4, 0, 1 + 2.0 * sin(a * M_PI / 180.0));
  glRotatef(a, 0, 1, 0);

  cube();
  glPopMatrix();

  glPushMatrix();
  glTranslatef(-1.4, 0, 1 + 2.0 * sin(a * M_PI / 180.0));
  glRotatef(-a, 0, 1, 0);

  cube();
  glPopMatrix();

  if(!pause)
    a += degres * dt / 1000.0;
  SDL_GL_SwapBuffers();
}

static void cube(void) {
  glBegin(GL_QUADS);

  /* Arriere */
  glNormal3f( 0,  0, -1);
  glTexCoord2f(1,  0);
  glVertex3f(-1,  1, -1);
  glTexCoord2f( 0,  0);
  glVertex3f( 1,  1, -1);
  glTexCoord2f( 0, 1);
  glVertex3f( 1, -1, -1);
  glTexCoord2f(1,  1);
  glVertex3f(-1, -1, -1);

  /* Gauche */
  glNormal3f(-1,  0,  0);
  glTexCoord2f(1,  1);
  glVertex3f(-1, -1,  1);
  glTexCoord2f(1,  0);
  glVertex3f(-1,  1,  1);
  glTexCoord2f(0,  0);
  glVertex3f(-1,  1, -1);
  glTexCoord2f(0,  1);
  glVertex3f(-1, -1, -1);

  /* Droite */
  glNormal3f( 1,  0,  0);
  glTexCoord2f(0,  1);
  glVertex3f( 1, -1,  1);
  glTexCoord2f(1,  1);
  glVertex3f( 1, -1, -1);
  glTexCoord2f(1,  0);
  glVertex3f( 1,  1, -1);
  glTexCoord2f(0,  0);
  glVertex3f( 1,  1,  1);

  /* Avant */
  glNormal3f( 0,  0,  1);
  glTexCoord2f(0,  0);
  glVertex3f(-1,  1,  1);
  glTexCoord2f(0,  1);
  glVertex3f(-1, -1,  1);
  glTexCoord2f(1,  1);
  glVertex3f( 1, -1,  1);
  glTexCoord2f(1,  0);
  glVertex3f( 1,  1,  1);

  /* Haut */
  glNormal3f( 0,  1,  0);
  glTexCoord2f(0,  1);
  glVertex3f(-1,  1,  1);
  glTexCoord2f(1,  1);
  glVertex3f( 1,  1,  1);
  glTexCoord2f(1,  0);
  glVertex3f( 1,  1, -1);
  glTexCoord2f(0,  0);
  glVertex3f(-1,  1, -1);

  /* Bas */
  glNormal3f( 0, -1,  0);
  glTexCoord2f(1,  1);
  glVertex3f(-1, -1,  1);
  glTexCoord2f(1,  0);
  glVertex3f(-1, -1, -1);
  glTexCoord2f(0,  0);
  glVertex3f( 1, -1, -1);
  glTexCoord2f(0,  1);
  glVertex3f( 1, -1,  1);

  glEnd();
}

/*
 * sdlFPS : compter le nombre de frames affiches. imprimer le
 * framerate toutes les 5 secondes.
 */
static inline void sdlFPS(void) {
  Uint32 t;
  static Uint32 t0 = 0, f = 0;
  f++;
  t = SDL_GetTicks();
  if(t - t0 > 5000) {
    fprintf(stderr, "sdlFPS:%6d\n",
	    (int)(1000.0 * f / (t - t0)));
    t0 = t;
    f  = 0;
  }
}

